@@ -9,9 +9,11 @@
 
 module GHC.Core.ConLike (
           ConLike(..)
+        , conLikeConLikeName
         , isVanillaConLike
         , conLikeArity
         , conLikeFieldLabels
+        , conLikeConInfo
         , conLikeInstOrigArgTys
         , conLikeUserTyVarBinders
         , conLikeExTyCoVars
@@ -29,19 +31,24 @@ module GHC.Core.ConLike (
 import GHC.Prelude
 
 import GHC.Core.DataCon
-import GHC.Core.PatSyn
-import GHC.Utils.Outputable
-import GHC.Types.Unique
-import GHC.Utils.Misc
-import GHC.Types.Name
-import GHC.Types.Basic
-import GHC.Core.TyCo.Rep (Type, ThetaType)
-import GHC.Types.Var
-import GHC.Core.Type(mkTyConApp)
 import GHC.Core.Multiplicity
+import GHC.Core.PatSyn
+import GHC.Core.TyCo.Rep (Type, ThetaType)
+import GHC.Core.TyCon (tyConDataCons)
+import GHC.Core.Type(mkTyConApp)
+import GHC.Types.Unique
+import GHC.Types.Name
+import GHC.Types.Name.Reader
+import GHC.Types.Basic
+
+import GHC.Types.GREInfo
+import GHC.Types.Var
+import GHC.Utils.Misc
+import GHC.Utils.Outputable
 
 import Data.Maybe( isJust )
 import qualified Data.Data as Data
+import qualified Data.List as List
 
 {-
 ************************************************************************
@@ -61,6 +68,10 @@ isVanillaConLike :: ConLike -> Bool
 isVanillaConLike (RealDataCon con) = isVanillaDataCon con
 isVanillaConLike (PatSynCon   ps ) = isVanillaPatSyn  ps
 
+conLikeConLikeName :: ConLike -> ConLikeName
+conLikeConLikeName (RealDataCon dc) = DataConName (dataConName dc)
+conLikeConLikeName (PatSynCon   ps) = PatSynName  (patSynName  ps)
+
 {-
 ************************************************************************
 *                                                                      *
@@ -113,6 +124,18 @@ conLikeFieldLabels :: ConLike -> [FieldLabel]
 conLikeFieldLabels (RealDataCon data_con) = dataConFieldLabels data_con
 conLikeFieldLabels (PatSynCon pat_syn)    = patSynFieldLabels pat_syn
 
+-- | The 'ConInfo' (arity and field labels) associated to a 'ConLike'.
+conLikeConInfo :: ConLike -> ConInfo
+conLikeConInfo con =
+  mkConInfo (conLikeConLikeInfo con) (conLikeArity con) (conLikeFieldLabels con)
+
+-- | Compute a 'ConLikeInfo' from a 'ConLike'.
+conLikeConLikeInfo :: ConLike -> ConLikeInfo
+conLikeConLikeInfo (RealDataCon con)
+  = ConIsData { conLikeDataCons = getName <$> tyConDataCons (dataConTyCon con) }
+conLikeConLikeInfo (PatSynCon {})
+  = ConIsPatSyn
+
 -- | Returns just the instantiated /value/ argument types of a 'ConLike',
 -- (excluding dictionary args)
 conLikeInstOrigArgTys :: ConLike -> [Type] -> [Scaled Type]
@@ -185,10 +208,14 @@ conLikeResTy (PatSynCon ps)    tys = patSynInstResTy ps tys
 --
 -- 7) The original result type
 conLikeFullSig :: ConLike
-               -> ([TyVar], [TyCoVar], [EqSpec]
+               -> ([TyVar], [TyCoVar]
                    -- Why tyvars for universal but tycovars for existential?
                    -- See Note [Existential coercion variables] in GHC.Core.DataCon
-                  , ThetaType, ThetaType, [Scaled Type], Type)
+                  , [EqSpec]
+                  , ThetaType      -- Provided theta
+                  , ThetaType      -- Required theta
+                  , [Scaled Type]  -- Arguments
+                  , Type )         -- Result
 conLikeFullSig (RealDataCon con) =
   let (univ_tvs, ex_tvs, eq_spec, theta, arg_tys, res_ty) = dataConFullSig con
   -- Required theta is empty as normal data cons require no additional
@@ -206,8 +233,10 @@ conLikeFieldType (RealDataCon dc) label = dataConFieldType dc label
 
 
 -- | The ConLikes that have *all* the given fields
-conLikesWithFields :: [ConLike] -> [FieldLabelString] -> [ConLike]
-conLikesWithFields con_likes lbls = filter has_flds con_likes
+conLikesWithFields :: [ConLike] -> [FieldLabelString]
+                   -> ( [ConLike]   -- ConLikes containing the fields
+                      , [ConLike] ) -- ConLikes not containing the fields
+conLikesWithFields con_likes lbls = List.partition has_flds con_likes
   where has_flds dc = all (has_fld dc) lbls
         has_fld dc lbl = any (\ fl -> flLabel fl == lbl) (conLikeFieldLabels dc)
 
