@@ -12,8 +12,7 @@ import Data.Binary
 import qualified Data.ByteString as B
 import qualified Data.ByteString.Internal as B
 import GHC.Serialized
-import qualified Language.Haskell.TH        as TH
-import qualified Language.Haskell.TH.Syntax as TH
+import qualified GHC.Internal.TH.Syntax as TH
 -- Put these in a separate module because they take ages to compile
 
 instance Binary TH.Loc
@@ -27,6 +26,7 @@ instance Binary TH.Info
 instance Binary TH.Type
 instance Binary TH.TyLit
 instance Binary TH.Specificity
+instance Binary TH.BndrVis
 instance Binary flag => Binary (TH.TyVarBndr flag)
 instance Binary TH.Role
 instance Binary TH.Lit
@@ -35,6 +35,7 @@ instance Binary TH.Stmt
 instance Binary TH.Pat
 instance Binary TH.Exp
 instance Binary TH.Dec
+instance Binary TH.NamespaceSpecifier
 instance Binary TH.Overlap
 instance Binary TH.DerivClause
 instance Binary TH.DerivStrategy
